/**
 * Javascript code for finding all tables with classname "striped" and
 * dynamically striping their row colors, and finding all tables with
 * classname "sortable" and making them dynamically sortable.
 *
 * TODO: incorporate missing features (if wanted) and improvements
 * from http://tetlaw.id.au/view/blog/table-sorting-with-prototype/
 * and http://www.millstream.com.au/view/code/tablekit/
 *
 * $Horde: horde/js/src/tables.js-prototype.js,v 1.6 2007/05/30 18:56:04 chuck Exp $
 *
 * See the enclosed file COPYING for license information (LGPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/lgpl.html.
 */

/* We do everything onload so that the entire document is present
 * before we start searching it for tables. */
Event.observe(window, 'load', table_init);

var SORT_COLUMN_INDEX;

function table_init()
{
    $$('TABLE.striped').each(table_stripe);
    $$('TABLE.sortable').each(table_makeSortable);
}

function table_stripe(table)
{
    var classes = ['rowEven', 'rowOdd'];

    // Tables can have more than one tbody element; get all child
    // tbody tags and interate through them.
    $A(table.childNodes).each(function(t) {
        if (t.tagName == 'TBODY') {
            $A(t.childNodes).each(function(c) {
                if (c.tagName == 'TR') {
                    Element.removeClassName(c, classes[1]);
                    Element.addClassName(c, classes[0]);
                    classes.reverse(true);
                }
            });
        }
    });
}

function table_makeSortable(table)
{
    var firstRow, i = 0;

    if (table.rows && table.rows.length > 0) {
        firstRow = table.rows[0];
    }
    if (!firstRow) {
        return;
    }

    // We have a first row: assume it's the header, and make its
    // contents clickable links.
    $A(firstRow.cells).each(function(cell) {
        if (Element.hasClassName(cell, 'nosort')) {
            ++i;
            return;
        }

        cell.columnIndex = i++;
        Element.setStyle(cell, { cursor: 'pointer' });
        Event.observe(cell, 'click', function(e) {
            if (e.target) {
                if (e.target.nodeType == 3) {
                    e.target = e.target.parentNode;
                }
            } else if (e.srcElement) {
                e.target = e.srcElement;
            }

            el = hasParent(e.target, 'A', 'TH');
            if (el && !Element.hasClassName(el, 'sortlink')) {
                return true;
            }

            table_resortTable(getParent(e.target, 'TH'));
            return false;
        });
    });
}

function table_getSortValue(el)
{
    var str = '';

    if (typeof el == 'string' ||
        typeof el == 'undefined') {
        return el;
    }

    // Use "sortval" if defined.
    if ((el.hasAttribute && el.hasAttribute('sortval')) ||
        typeof el['sortval'] != 'undefined') {
        return el.getAttribute('sortval');
    }

    if (el.innerText) {
        // Not needed but it is faster.
        return el.innerText;
    }

    $A(el.childNodes).each(function(e) {
        switch (e.nodeType) {
        case 1:
            // ELEMENT_NODE
            str += table_getSortValue(e);
            break;

        case 3:
            // TEXT_NODE
            str += e.nodeValue;
            break;
        }
    });

    return str;
}

function table_resortTable(th)
{
    var table = getParent(th, 'TABLE'), sortColumn = th.columnIndex, sortDown = 0;

    // Loop through <thead> to find the current sort column and
    // direction.
    $$('TH').each(function(e) {
        if (th == e) {
            if (Element.hasClassName(e, 'sortup')) {
                Element.removeClassName(e, 'sortup');
                Element.addClassName(e, 'sortdown');
            } else if (Element.hasClassName(e, 'sortdown')) {
                Element.removeClassName(e, 'sortdown');
                Element.addClassName(e, 'sortup');
                sortDown = 1;
            } else {
                Element.addClassName(e, 'sortdown');
            }
        } else {
            Element.removeClassName(e, 'sortup');
            Element.removeClassName(e, 'sortdown');
        }
    });

    // Work out a type for the column
    if (table.rows.length <= 1) {
        return;
    }

    var itm = table_getSortValue(table.rows[1].cells[sortColumn]);
    var sortfn = table_sort_caseinsensitive;
    if (itm.match(/^\d\d[\/-]\d\d[\/-]\d\d\d\d$/) ||
        itm.match(/^\d\d[\/-]\d\d[\/-]\d\d$/)) {
        sortfn = table_sort_date;
    } else if (itm.match(/^[$]/)) {
        sortfn = table_sort_currency;
    } else if (itm.match(/^[\d\.]+$/)) {
        sortfn = table_sort_numeric;
    }

    SORT_COLUMN_INDEX = sortColumn;

    // Don't mix up seperate tbodies; sort each in turn.
    for (var i = 0; i < table.tBodies.length; ++i) {
        trs = table.tBodies[i].getElementsByTagName('tr');
        newRows = [];
        for (var j = 0; j < trs.length; ++j) {
            newRows[j] = trs[j];
        }

        newRows.sort(sortfn);
        if (sortDown) {
            newRows.reverse();
        }

        // We appendChild rows that already exist to the tbody, so it
        // moves them rather than creating new ones. Don't do
        // sortbottom rows.
        for (var j = 0; j < newRows.length; ++j) {
            if (!Element.hasClassName(newRows[j], 'sortbottom')) {
                table.tBodies[i].appendChild(newRows[j]);
            }
        }

        // Do sortbottom rows only.
        for (var j = 0; j < newRows.length; ++j) {
            if (Element.hasClassName(newRows[j], 'sortbottom')) {
                table.tBodies[i].appendChild(newRows[j]);
            }
        }
    }

    // If we just resorted a striped table, re-stripe it.
    if (Element.hasClassName(table, 'striped')) {
        table_stripe(table);
    }

    // Finally, see if we have a callback function to trigger.
    if (typeof(table_sortCallback) == 'function') {
        table_sortCallback(table.id, th.id, sortDown);
    }
}

function getParent(el, pTagName)
{
    if (el == null) {
        return null;
    } else if (pTagName == null) {
        return el.parentNode;
    } else if (el.nodeType == 1 && el.tagName.toLowerCase() == pTagName.toLowerCase()) {
        // Gecko bug, supposed to be uppercase.
        return el;
    } else {
        return getParent(el.parentNode, pTagName);
    }
}

function table_sort_date(a, b)
{
    // Two digit years less than 50 are treated as 20XX, greater than
    // 50 are treated as 19XX.
    var aa = table_getSortValue(a.cells[SORT_COLUMN_INDEX]),
        bb = table_getSortValue(b.cells[SORT_COLUMN_INDEX]),
        dt1, dt2, yr;

    if (aa.length == 10) {
        dt1 = aa.substr(6, 4) + aa.substr(3, 2) + aa.substr(0, 2);
    } else {
        yr = aa.substr(6, 2);
        if (parseInt(yr) < 50) {
            yr = '20' + yr;
        } else {
            yr = '19' + yr;
        }
        dt1 = yr + aa.substr(3, 2) + aa.substr(0, 2);
    }
    if (bb.length == 10) {
        dt2 = bb.substr(6, 4) + bb.substr(3, 2) + bb.substr(0, 2);
    } else {
        yr = bb.substr(6, 2);
        if (parseInt(yr) < 50) {
            yr = '20' + yr;
        } else {
            yr = '19' + yr;
        }
        dt2 = yr + bb.substr(3, 2) + bb.substr(0, 2);
    }
    if (dt1 == dt2) {
        return 0;
    } else if (dt1 < dt2) {
        return -1;
    }
    return 1;
}

function table_sort_currency(a, b)
{
    var aa = table_getSortValue(a.cells[SORT_COLUMN_INDEX]).replace(/[^0-9.]/g, ''),
        bb = table_getSortValue(b.cells[SORT_COLUMN_INDEX]).replace(/[^0-9.]/g, '');
    return parseFloat(aa) - parseFloat(bb);
}

function table_sort_numeric(a, b)
{
    var aa = parseFloat(table_getSortValue(a.cells[SORT_COLUMN_INDEX]));
    if (isNaN(aa)) {
        aa = 0;
    }
    var bb = parseFloat(table_getSortValue(b.cells[SORT_COLUMN_INDEX]));
    if (isNaN(bb)) {
        bb = 0;
    }
    return aa - bb;
}

function table_sort_caseinsensitive(a, b)
{
    var aa = table_getSortValue(a.cells[SORT_COLUMN_INDEX]).toLowerCase(),
        bb = table_getSortValue(b.cells[SORT_COLUMN_INDEX]).toLowerCase();
    if (aa == bb) {
        return 0;
    } else if (aa < bb) {
        return -1;
    }
    return 1;
}

function table_sort_default(a, b)
{
    var aa = table_getSortValue(a.cells[SORT_COLUMN_INDEX]),
        bb = table_getSortValue(b.cells[SORT_COLUMN_INDEX]);
    if (aa == bb) {
        return 0;
    } else if (aa < bb) {
        return -1;
    }
    return 1;
}

/**
 * DOM utility functions.
 */
function hasParent(el, tagName, tagStop)
{
    if (el.tagName == tagName) {
        return el;
    } else if (tagStop != null && el.tagName == tagStop) {
        return false;
    } else {
        return hasParent(getParent(el), tagName, tagStop);
    }
}
